'''*****************************************************************************
AToMPM - A Tool for Multi-Paradigm Modelling

Copyright (c) 2011 Raphael Mannadiar (raphael.mannadiar@mail.mcgill.ca)

This file is part of AToMPM.

AToMPM is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later 
version.

AToMPM is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with AToMPM.  If not, see <http://www.gnu.org/licenses/>.
*****************************************************************************'''

import re, json, uuid, cPickle as pickle, threading, itertools, traceback, logging
from utils import Utilities as utils
from tcontext import ModelTransformationContext, ExhaustContext
from tconstants import TConstants as TC
from compiler import ModelAndRuleCompiler
from pytcore.core.himesis import HConstants as HC
from pytcore.rules.arule import ARule
from pytcore.tcore.messages import Packet
import igraph as ig
import __builtin__ 
import pydot
import datetime
from random import choice
from threading import *
from Queue import *
from barrier import *
from synchgraph import *
from itertools import *
from petrinet import *
from _abcoll import Iterable
from time import *
import timeit
from pprint import isreadable
''' 
	py-t-core abstraction layer 
	
	_lock		  						used to synchronize access to self._changelogs
	_changelogs 					used to hold asworker changelogs
	_M			  						the himesis graph we're transforming
	_mtContexts						the current transformation contexts
	_transfData						a mapping between transformation filenames and 
										their data 	
	_userTransfs					stores the transformations loaded by the user 
										(needed to return to the pre-run state after 
										running)
	_execmode						the current execution mode 
	_aswHttpReq						pointer to a function that sends HTTP requests 
										to our parent mtworker's asworker
	_aswNextSequenceNumber		used to determine if a changelog is received out
  										of order, and if a pending changelog is now ready
									  	to be handled
	username							the user's username '''
class PyTCoreAbstractionLayer :
	def __init__(self,aswHttpReq) :
		self._lock 	  					 = threading.Condition()
		self._changelogs 				 = []
		self._M			  				 = None
		
		'''
		Used only in COMP 522 and comp 621
		Petri Net Modules, docomposition of disconnected
		_M graph
		'''
		self.modules						= {}
		'''
		Synchronization graph for modular
		state spaces
		'''
		self.SG = None
		self.modStart = None
		self.modEnd = None
		self.flatStart = None
		self.flatEnd = None
		'''
		State spaces for individual modules
		'''
		self.modIDtoPOS= {}
		self.SSs = {}
		'''
		'''
		self._mtContexts				 = []
		self._transfData				 = {}
		self._userTransfs 			 = []
		self._execmode					 = 'STOPPED'
		self._aswHttpReq				 = aswHttpReq
		self._aswNextSequenceNumber = None
		self.username					 = None
		

	'''
		send a PUT /GET/console request to the asworker '''
	def _aswPrintReq(self,msg) :
		logging.debug('CONSOLE :: '+msg)
		self._aswHttpReq('PUT','/GET/console',{'text':msg})
		
	
	'''
		do 'callback()' once we've received the feedback for the last step that
	  	ran, if any, checking for the said feedback and handling any newly 
		received feedback	every TC.WAIT_ON_CHLOG_DELAY seconds '''
	def _doWhenLastStepFeedbackReceived(self,callback) :
		def condition() :
			self._handleChangelogs()
			return self._mtContexts[-1].isLastStepFeedbackReceived()
		utils.doWhen(
			condition,
			TC.WAIT_ON_CHLOG_DELAY,
			callback)


	'''
		synchronously verify if any changelogs have been received, and if we are 
		ready to handle them, remove them from self._changelogs and handle them

		1. sort changelogs by ascending sequence#
		2. do nothing if no self._changelogs is empty, if no model has been loaded
	  		yet (i.e., self._aswNextSequenceNumber == None), or if the oldest 
			changelog is still too new to be handled
		3. crash if we encouter an invalid sequence#
		4. otherwise (we're ready to handle the oldest changelog), handle the 
			oldest changelog, increment self._aswNextSequenceNumber and recurse
			back to step 1

		TBI: performance would benefit greatly from caches that map atompm ids to
			  GUIDs '''
	def _handleChangelogs(self) :
		
		'''
			handle a single changelog '''
		def _handleChangelog(changelog) :
			def eq(a,b) : return str(a) == str(b)

			for c in changelog :
				if c['op'] == 'MKEDGE' : 
					node1 = \
						self._M.vs.select(lambda v : eq(v['$atompmId'],c['id1']))[0]
					node2 = \
						self._M.vs.select(lambda v : eq(v['$atompmId'],c['id2']))[0]
					self._M.add_edges((node1.index, node2.index))
				
				elif c['op'] == 'RMEDGE' : 
					pass
	
				elif c['op'] == 'MKNODE' :
					self._compiler.addNode(self._M, json.loads(c['node']), c['id'])
	
				elif c['op'] == 'RMNODE' :
					node = \
						self._M.vs.select(lambda v : eq(v['$atompmId'],c['id']))[0]
					self._M.delete_nodes([node.index]) 
	
				elif c['op'] == 'CHATTR' : 
					node = \
						self._M.vs.select(lambda v : eq(v['$atompmId'],c['id']))[0]
					self._M.vs[node.index][c['attr']] = c['new_val']
					
				elif c['op'] == 'LOADMM' : 
					self._compiler.parseMetamodel(
							c['name'],
							utils.fread(
								'/users/%s/%s.metamodel'%(self.username,c['name'])),
							loadMM=True)
	
				elif c['op'] == 'DUMPMM' : 
					self._compiler.unloadMetamodel(c['name'])
	
				elif c['op'] == 'RESETM' : 
					self._M = self._compiler.compileModel(c['new_model'])
					self._M.mmTypeData = self._compiler.getMMTypeData()
	
				elif c['op'] == 'SYSOUT' : 
					if c['text'] == TC.RULE_SUCCESS_MSG or \
						c['text'] == TC.RULE_NOT_APPLICABLE_MSG or \
						c['text'].startswith(TC.RULE_FAILURE_MSG) or \
						c['text'].startswith(TC.TRANSFORMATION_DONE) or \
						c['text'].startswith(TC.REMOTE_APPLICATION_FAILURE) :
						self._mtContexts[-1].setLastStepFeedbackReceived()


		self._lock.acquire()
		self._changelogs.sort(key=lambda c : utils.sn2int(c['sequence#']))
		if len(self._changelogs) == 0 or \
			self._aswNextSequenceNumber == None or \
			utils.sn2int(self._changelogs[0]['sequence#']) > \
				utils.sn2int(self._aswNextSequenceNumber) :
			self._lock.release()
		else : 
			sn = self._changelogs[0]['sequence#']
			if utils.sn2int(sn) < utils.sn2int(self._aswNextSequenceNumber) :
				raise ValueError('invalid changelog sequence# :: '+sn)
			else :
				logging.debug('++ ('+sn+') '+str(self._changelogs[0]['changelog']))
				_handleChangelog(self._changelogs.pop(0)['changelog'])
				self._aswNextSequenceNumber = \
							utils.incrementSequenceNumber(self._aswNextSequenceNumber)
				self._lock.release()
				self._handleChangelogs()
	


	'''
		load a model (and its metamodels)

		1. compile the provided model into a himesis graph and save it to self._M
		2. synchronize self._M's mmTypeData with that of self._compiler's so that
			it gets updated as new metamodels are loaded
		3. initialize self._aswNextSequenceNumber based on the sequence# 'sn' of
			the provided model, and forget any already received out-of-date 
			changelogs

		NOTE: this function should only get called once (when the asworker 
				initially sets up this mtworker) '''
	def loadModel(self,m,mms,sn) :
		assert self._M == None, 'ptcal.loadModel() should only be called once'
		self._compiler = ModelAndRuleCompiler(self.username)
		self._M = self._compiler.compileModel(m,mmsData=mms)
		print 'load model'
		self.modules = {}
		self._M.mmTypeData = self._compiler.getMMTypeData()
		self._aswNextSequenceNumber = utils.incrementSequenceNumber(sn)
		self._lock.acquire()
		self._changelogs = \
			filter(lambda c : utils.sn2int(sn) < utils.sn2int(c['sequence#']), 
					 self._changelogs)
		self._lock.release()
		
		
	'''
		load a PN model (and its metamodels)

		1. compile the provided model into a himesis graph and save it to self._M
		2. synchronize self._M's mmTypeData with that of self._compiler's so that
			it gets updated as new metamodels are loaded
		3. initialize self._aswNextSequenceNumber based on the sequence# 'sn' of
			the provided model, and forget any already received out-of-date 
			changelogs

		NOTE: this function should only get called once (when the asworker 
				initially sets up this mtworker) '''
	def loadModelPN(self,m,mms,sn) :
		assert self._M == None, 'ptcal.loadModel() should only be called once'
		self._compiler = ModelAndRuleCompiler(self.username)
		self._M = self._compiler.compileModelPN(m,mmsData=mms)
		disjoint = self._M.decompose(mode=ig.WEAK)
		print 'LoADING pn MODEL'
		#create dictionary of modules
		for mod in disjoint:
			self.modules [ uuid.uuid4()] = mod
		self._M.mmTypeData = self._compiler.getMMTypeData()
		self._aswNextSequenceNumber = utils.incrementSequenceNumber(sn)
		self._lock.acquire()
		self._changelogs = \
			filter(lambda c : utils.sn2int(sn) < utils.sn2int(c['sequence#']), 
					 self._changelogs)
		self._lock.release()
		
	

	''' setup internal state to reflect given runtime configuration '''
	def _loadRuntimeConfiguration(self,rc) :
		if 'looseSubtypingMM' in rc :
			self._compiler.RC__looseSubtypingMM = rc['looseSubtypingMM'] 



	'''
		read in some json that describes a model transformation from file 'fname',
		store in self._transfData, and push a new ModelTransformationContext onto
		self._mtContexts '''
	def _loadTransform(self,fname) :
		if fname not in self._transfData :
			self._transfData[fname] = \
				utils.fread('/users/%s/%s'%(self.username,fname))
		self._mtContexts.append( 
				ModelTransformationContext(self._transfData[fname]) )



	'''
		load a set of user-specified transformations and forget anything 
		previously loaded
		
		1. forget previously loaded transformations and compiled rules 
		2. load transformations
		3. remember loaded transformations '''
	def loadTransforms(self,fnames) :
		self._transfData	= {}
		self._mtContexts	= []
		self._compiler.forgetCompiledRules()
		self._M.session	= {}
    
		for fname in fnames :
			self._loadTransform(fname)
		self._userTransfs = fnames



	''' 
		returns the filename and compiled form of the next rule to run
		
		1. fetch the current transformation context
		2. 
		3. call its nextStep() function to get the next transformation step
			a) if the step is not an object (i.e., is a application code)
				i. if there is more than 1 loaded transformation context, 
					j.   pop the newly completed transformation context
					jj.  if the next context (which may be a "parent" transformation
						  or the next of multiple loaded transformations) is a parent,
						  update its last step application info
  				   jjj. make a recursive call to get the next rule to run from 
						  within the "next" context
				ii. otherwise, simply return the said application code				
  			b) if the step is an error, return it
			c) otherwise, determine the type of the step (via the node's $type 
				attribute or by inspecting the 'metamodels' array in the provided 
				.model file) 
					i.  Rules get stored in 'nr' to be later returned 
				  	ii. Transformations and Exhausts[Randoms] cause the pushing of a
				  		 new transformation context onto self._mtContexts, and of a
					  	 recursive call to get the next rule to run from within that
						 new context	  
		4. return the rule name and compiled form ''' 
	def _nextRule(self) :
		mtc = self._mtContexts[-1]
		self._loadRuntimeConfiguration(mtc.getRuntimeConfiguration())
		ns = mtc.nextStep()

		if ns.__class__ != {}.__class__ :
			if len(self._mtContexts) > 1 :
				self._mtContexts = self._mtContexts[:-1]
				if self._mtContexts[-1].isTransformationUnderWay() :
					self._mtContexts[-1].setLastStepApplicationInfo(ns)
				return self._nextRule()
			else :
				return ns

		elif '$err' in ns :
			return ns['$err']

		else :
			if 'id' in ns :
				fulltype = mtc.t['nodes'][ns['id']]['$type']
				mm = '/Formalisms/Transformations.../Transformation/Transformation/'
				if fulltype == mm+'Rule' :
					return {'fname':ns['fname'],
	 						  'cr':self._compiler.compileRule(None,ns['fname'])}

				elif fulltype == mm+'Transformation' :
					self._loadTransform(ns['fname'])
					return self._nextRule()

				elif fulltype == mm+'Exhaust' :
					self._mtContexts.append( ExhaustContext(mtc.t,ns['id']) )
					return self._nextRule()

				elif fulltype == mm+'ExhaustRandom' :
					self._mtContexts.append( ExhaustContext(mtc.t,ns['id'],True) )
					return self._nextRule()
			else :
				rulemm	= '/Formalisms/Transformations.../TransformationRule/'+\
 							  'TransformationRule'
				transfmm = '/Formalisms/Transformations.../Transformation/'+\
 							  'Transformation'
				contents = utils.fread('/users/%s/%s'%(self.username,ns['fname']))

				if rulemm in contents['metamodels'] :
					return {'fname':ns['fname'],
	 						  'cr':self._compiler.compileRule(contents,ns['fname'])}

				elif transfmm in contents['metamodels'] :
					self._transfData[ns['fname']] = contents
					self._loadTransform(ns['fname'])
					return self._nextRule()

				raise ValueError(\
							'file does not contain valid rule or transformation '+\
							'model :: '+ns['fname'])
		


	'''
		synchronously save 1 changelog into self._changelogs '''
	def onchangelog(self,c) :
		self._lock.acquire()
		self._changelogs.append(c)
		self._lock.release()



	'''
		causes the execution of the current transformation(s) to pause (by 
		preventing _play()'s next call to _step(), if any) '''
	def pause(self) :
		#__builtin__.profileLog.dump()
		self._execmode = 'PAUSE'



	'''
		play()
		calls _play() if it isn't already running (i.e., if we're already in PLAY
		mode, a timed call to _play() has already been placed)


		_play()
		schedules an action for when feedback from the last step is received... 
		the action is
			1. return if we're no longer in PLAY mode
			2. take one _step()
			3. schedule a recursive call to _play() in TC.INTER_RULE_DELAY 
				seconds '''
	def play(self) :
		if self._execmode != 'PLAY' :
			self._execmode = 'PLAY'
			self._play()		
	def _play(self) :
		self._doWhenLastStepFeedbackReceived(
					lambda : self._execmode == 'PLAY' and \
								self._step() and \
								utils.setTimeout(TC.INTER_RULE_DELAY,self._play))



	'''
		step()
		wrapper around _step() that ensures that step requests from user are 
		ignored when in PLAY mode, and that valid requests only go through (i.e.,
		actually call _step()) when feedback for the last step is received


		_step()
		fetch and run next rule

		1. fetch next rule
			a) if next rule is not a {} (i.e., all available transformations have
			  	terminated and _nextRule() returned the resulting application code),
				report application code and stop()
			b) if an error is returned, report it and stop()
			c) otherwise, 
				i.   run rule (returns (deltas|error,applicationInfo))
				ii.  set ran rule's application info
				iii. if rule was n/a, report this
				iii. if rule failed, report this and error
				iii. otherwise,
					j.   construct a batchEdit operation based on the rule's effects,
						  and unfo the said effects
					jj.  send off the batchEdit
					jjj. stop() if the batchEdit fails 

		NOTE: this function assumes that feedback for the last step has already
				been received '''
	def step(self) :
		if self._execmode == 'PLAY' :
			pass
		else : 
		 	self._execmode = 'STEP'
		 	self._doWhenLastStepFeedbackReceived(self._step)
	def _step(self) :		

		'''
			go through a rule's deltas and (1) produce a batchEdit request, and 
			(2) undo them

			NOTE: we undo the deltas for 2 reasons
					1. so that changes become driven by asworker changelogs (like in 
						csworkers)
					2. so that we don't need to figure out which entries in received
						changelogs correspond to user-operations and which ones 
						correspond to the effects of the constructed batchEdit 
			NOTE: since we sometimes need to use the result from one request as the
					parameter of another (i.e., create a node, update *it*), we use
					the 'mknodes' map to remember which requests created which new 
					nodes... this also allows to know which nodes already exist and
				  	which ones were created by the last rule 
			NOTE: because creation of connector nodes and their linking to their 
					ends is described in non-contiguous deltas, we use the 'mknodes'
					map to remember incomplete connector creation requests until the
					appropriate MKEDGE deltas are encountered '''
		def buildEditHttpReq(deltas) :
			reqs    		 = []
			mknodes 		 = {}
			neighborhood = None

			'''
				construct an atompmId given a node... the result will be
			  	a) a 'macro' to be replaced by the result of an earlier request 
					within the batchEdit, if the node was created by the last rule
				b) the atompmId stored within the node, if the node already has 
					a counter-part in atompm '''
			def atompmInstanceId(node) :
				if node[HC.GUID] in mknodes :
					return '$'+str(mknodes[node[HC.GUID]])+'$'
				else :
					return node['$atompmId']


			for d in deltas :
				if d['op'] == 'RMNODE' :
					reqs.append({\
							'method':'DELETE',
							'uri':d['attrs'][HC.FULLTYPE]+'/'+\
									d['attrs']['$atompmId']+'.instance'})

				elif d['op'] == 'MKNODE' :
					mknodes[d['guid']] = len(reqs)
					node = self._M.vs[self._M.get_node(d['guid'])]
					if neighborhood == None :
						neighborhood = map(
							lambda n: n[HC.FULLTYPE]+'/'+n['$atompmId']+'.instance',
							d['neighborhood'])				
					if node[HC.CONNECTOR_TYPE] :
						reqs.append({\
								'method':'POST',
								'uri':node[HC.FULLTYPE]+'.type',
								'reqData':
									{'src':None,
									 'dest':None,
									 'hitchhiker':
									 	{'segments':None,
										 'asSrc':None,
										 'asDest':None,
										 'neighborhood':neighborhood}}})	
					else :
						reqs.append({\
								'method':'POST',
								'uri':node[HC.FULLTYPE]+'.type',
								'reqData':{'hitchhiker':{'neighborhood':neighborhood}}})

				elif d['op'] == 'RMEDGE' :
					pass

				elif d['op'] == 'MKEDGE' :
					def isConnectorMKNODE(req): 
						return 'dest' in req['reqData']

					if d['guid1'] in mknodes :
						req = reqs[ mknodes[d['guid1']] ]
						if isConnectorMKNODE(req) :
							node2 = self._M.vs[self._M.get_node(d['guid2'])]
							id		= atompmInstanceId(node2)
							req['reqData']['dest'] = \
							req['reqData']['hitchhiker']['asDest'] = \
								node2[HC.FULLTYPE]+'/'+id+'.instance'

					if d['guid2'] in mknodes :
						req = reqs[ mknodes[d['guid2']] ]
						if isConnectorMKNODE(req) :
							node1 = self._M.vs[self._M.get_node(d['guid1'])]
							id		= atompmInstanceId(node1)
							req['reqData']['src'] = \
							req['reqData']['hitchhiker']['asSrc'] = \
								node1[HC.FULLTYPE]+'/'+id+'.instance'
						
				elif d['op'] == 'CHATTR' :
					node = self._M.vs[self._M.get_node(d['guid'])]				
					id   = atompmInstanceId(node)
					reqs.append({\
							'method':'PUT',
							'uri':node[HC.FULLTYPE]+'/'+id+'.instance',
							'reqData':{'changes':{d['attr']:d['new_val']}}})

				elif d['op'] == 'LOADMM' :
					reqs.append({\
						'method':'PUT',
						'uri':'/current.metamodels',
						'reqData':
							{'mm':'/%s%s.metamodel'%(self.username,d['name'])}})
					

			for d in reversed(deltas) :
				if d['op'] == 'RMNODE' :
					newNodeIndex = self._M.add_node(newNodeGuid=d['attrs'][HC.GUID])
					for attr,val in d['attrs'].iteritems() :
						self._M.vs[newNodeIndex][attr] = val

				elif d['op'] == 'MKNODE' :
					node = self._M.vs[self._M.get_node(d['guid'])]				
					self._M.delete_nodes([node.index])

				elif d['op'] == 'RMEDGE' :
					node1 = self._M.vs[self._M.get_node(d['guid1'])]				
					node2 = self._M.vs[self._M.get_node(d['guid2'])]				
					self._M.add_edges((node1.index, node2.index))

				elif d['op'] == 'MKEDGE' :
					pass
						
				elif d['op'] == 'CHATTR' :
					node = self._M.vs[self._M.get_node(d['guid'])]
					node[d['attr']] = d['old_val']

				elif d['op'] == 'LOADMM' :
					pass


			reqs.append({\
					'method':'PUT',
					'uri':'/GET/console',
					'reqData':{'text':TC.RULE_SUCCESS_MSG}})
			return reqs


		'''
			run the specified rule and return a tuple describing its execution '''
		def runRule(r) :
			self._aswPrintReq('launching rule :: '+r['fname'])
 			ar = ARule(r['cr']['lhs'],r['cr']['rhs'])
			packet = ar.packet_in( Packet(self._M) )
			if ar.is_success :
				return (packet.deltas,TC.SUCCEEDED)
			elif not ar.is_success :
				if ar.exception :
					return (str(ar.exception),TC.FAILED)
				else :
					return (None,TC.NOT_APPLICABLE)


		try :
			nr = self._nextRule()
		except Exception :
			nr = {'$err':traceback.format_exc()}

		if nr.__class__ != {}.__class__ :
			self._aswPrintReq(TC.TRANSFORMATION_DONE + nr)
			__builtin__.profileLog.dump()
			self.stop()
			return

		elif '$err' in nr :
			self._aswPrintReq(TC.NO_NEXT_RULE+nr['$err'])
			self._stop()
			return

		else :
			(res,ai) = runRule(nr)
			self._mtContexts[-1].setLastStepApplicationInfo(ai)

			if ai == TC.NOT_APPLICABLE :
				self._aswPrintReq(TC.RULE_NOT_APPLICABLE_MSG)
			elif ai == TC.FAILED :
				self._aswPrintReq(TC.RULE_FAILURE_MSG + res)
			else : 
				req  = buildEditHttpReq(res)
				resp = self._aswHttpReq('POST','/batchEdit',req)
				if not utils.isHttpSuccessCode(resp['statusCode']) :
					self.stop()
					self._aswPrintReq(TC.REMOTE_APPLICATION_FAILURE + resp['reason'])
					return

		return True



	'''
		stop()
		sets self._execmode to STOPPING and schedules a call to _stop() for when
		feedback from the last step is received... being in STOPPING mode implies
		that the _play() loop, if any, will be broken, and that incoming user 
		requests will be rejected


		_stop()
		1. restores self._mtContexts to right after the user loaded his 
			transformation(s) (i.e., to before we actually [partially] ran it)
		2. sends a console message to notify the user that the transformation has
			stopped
		3. sets self._execmode to STOPPED (i.e., we're done STOPPING and can 
			handle new requests) '''
	def isStopped(self)  :	return self._execmode == 'STOPPED'
	def isStopping(self) :	return self._execmode == 'STOPPING'
	def stop(self) :
		self._execmode = 'STOPPING'
		self._doWhenLastStepFeedbackReceived(self._stop)
	def _stop(self) :
		self._mtContexts = []
		for fname in self._userTransfs :
			self._loadTransform(fname)
		self._aswPrintReq(TC.TRANSFORMATION_STOPPED)
		__builtin__.profileLog.dump()
		self._execmode = 'STOPPED'
	
	def createMarking(self, net):
		marking = {}
		places = net.vs.select(type_eq = 'P')
		for p in places:
			marking[p['name']] = int(p['nbTokens'])
		return marking
	
#	#Non infinite reachability. for now 
#	#actually this is a simulator!!!!!!!!!!!!!!
#	def reachability(self,key,net):
#		#self.SSs[key] = self.createMarking(net)
#		g = ig.Graph(1,directed=True)
#		#g.add_vertices(1) #add first node
#		g.vs[g.vcount()-1]['M'] = self.createMarking(net) #add M0
#		print 'Adding initial state %s'%(g.vs[g.vcount()-1]['M'])
#		transitions = net.vs.select(type_eq = 'T')
#		lastmarking = g.vcount()-1
#		run = True
#		while run:
#			run = False
#			enabled = self.getEnabled(transitions,net)
#			t = self.firetrans(enabled)
#			if t: #fire transition 
#				print 'Transition %s fires'%(t['name'])
#				run = True
#				Pin = net.predecessors(t.index)
#				Pout = net.successors(t.index) #get outgoing
#				for p in Pin:
#					nbTokens = int(net.vs[p]['nbTokens'])
#					if nbTokens >0:
#						net.vs[p]['nbTokens'] = nbTokens - 1 #ideally we want to use the weight of the arc... 1 
#				for p in Pout:
#					nbTokens = int(net.vs[p]['nbTokens'])
#					net.vs[p]['nbTokens'] = nbTokens + 1 # add  token to each out place
#				#need to add new state to graph and draw transition ark
#				newstate = self.createMarking(net)
#				print 'New state %s'%(newstate)
#				exists = False
#				id = None
#				for v in g.vs:
#					if v['M'] == newstate:
#						exists = True
#						id = v.index
#						break
#				if exists: #marking exists add edge to iself and add transition name as attribute to an edge
#					print 'Marking is already present edge %s->%s'%(lastmarking,id)
#					g.add_edges([(lastmarking,id)])
#					g.es[g.get_eid(lastmarking, id)]['T'] = t['name']
#					lastmarking = id
#					#print 'last id %s'%(newid)
#					#break
#				else:
#					#print 'Creating new marking'
#					g.add_vertices(1)
#					newid = g.vcount()-1
#					print 'new id edge %s->%s'%(lastmarking,newid)
#					g.vs[newid]['M'] = newstate
#					g.add_edges([(lastmarking,newid)])
#					g.es[g.get_eid(lastmarking, newid)]['T'] = t['name']
#					lastmarking = newid
#					#break
#				continue
#		return g
	
		#First create marking dictionary from places
		#Iterate through transitions and see if they are enabled
		#Do we find all transitions and choose transition to fire
		#or we fire at once as we find the transition.
		#change produce new marking and update the graph accordingly.
	#//COMP 522 project start		
	def getEnabled(self,transitions,net):
		enabled = []
		for t in transitions:
			en = True
			Pin = net.predecessors(t.index) #get incoming places into transitions
			for p in Pin:
				if not int(net.vs[p]['nbTokens']) >= 1: #weight of arcs is 1 for the time being, will need to replace to weingtfrom the arc
					en = False
					break
			if en:
				enabled.append(t)
		return enabled
	
	#Fire transition for execution
	def firetrans(self,enabled):
		if enabled:
			return choice(enabled)
		else: 
			return None
	
	#Perform flat reachability analysis
	def PNFull(self):
		#if not self.modules:
		self.modules = {}		
		self.modStart = time()
		disjoint = self._M.decompose(mode=ig.WEAK)
		'''Let's compile PN graph out of generic himesis graph and create dictionary with unique IDs'''
		barier = barrier(len(disjoint)+1) #barier, modules + synch thread
		#queue = Queue()
		for mod in disjoint:
			queue = Queue()
			uid = uuid.uuid4()
			key = str(uid)[:str(uid).index('-')]
			#self.modules [ uid] = self.toPN(mod)
			module = PnModule(self.toPN(mod),queue, barier,True)
			self.modules [ key] = module
			module.start()
		barier.wait()
		print 'Time elapsed flat analysis %f'%(time() - self.modStart)
		self.modules [ key].summary()
		#self.reachabilityTest()
		for key,mod in self.modules.items():
			mod.graph()
	
	#peroform standard reachability analysis
	def analyzePN(self):
		'''First lets break model into submodels'''
		self.modStart = time()
		#if not self.modules:
		self.modules = {}
		disjoint = self._M.decompose(mode=ig.WEAK)
		barier = barrier(len(disjoint)+1) #barier, modules + synch thread
		for mod in disjoint:
			queue = Queue()
			module = PnModule(self.toPN(mod),queue, barier)
			self.modules [ module.getKey()] = module
			module.start()
		barier.wait() #wait till all threads stop doing first phase.
		M0 = []
		TFenabled = []
		TF = {}
		work = [] 
		tofire = []
		for key,mod in self.modules.items():
			TF[key] = mod.TFS
			M0.append('%s-%d'%(key,0))
		ind=0
		self.sg = synchgraph(len(self.modules),M0)
		work.append(M0)	
		res={}
		while work:
			M = work.pop()
			for key,mod in self.modules.items():
				TFenabled.append( mod.getEnabledTFs())
			tofire = reduce(set.intersection,map(set,TFenabled)) 	
			for key,mod in self.modules.items():
				mod.que.put(tofire)
			barier.wait() #lets wait for threads, they may building local graphs still
			end = False
			for key,mod in self.modules.items():
				if not mod.result:
					end = True
				res[key] = mod.result #got results now produce new states for syngraph and archs.
			if not end:
				work.append(M)
			else:
				#self.sg.graph()
				for key,mod in self.modules.items():
					mod.SC()
					mod.graph()
					mod.que.put(['@exit'])
				self.sg.markSCC(self.modules)
				self.sg.graph()
				print '---------------------------'
				print 'Time elapsed modular analysis %f'%(time() - self.modStart)
				for key,mod in self.modules.items():
					mod.summary()
					print '---------------------------'
				print 'Synch graph:'
				self.sg.summary()
				print '---------------------------'
				self.modStart = time()
				self.reachabilityTestModular()
				print 'Find place elapsed modular analysis %f'%(time() - self.modStart)
				return
			#main result
			fr ={}
			to = {}
			for key,value in res.items():
				for k,v in value.items():
					if not k in fr:
						fr[k] = []
						fr[k].append([])
						fr[k].append([])
					fr[k][0].append(v[0])
					fr[k][1].append(v[1])
			from_prod=[]
			to_prod = []
			T=None
			for key,value in fr.items():
				T = key
				#res = list(list(itertools.product(*value[0]))[0])
				from_prod.append(list( list(itertools.product(*value[0]))[0]))
				to_prod.append(list (list(itertools.product	(*value[1]))[0]))
			self.sg.addMarkingBatch(T,from_prod,to_prod)
			#	self.sg.addMarking(from_prod[i],to_prod[i],T)
			res.clear()
			TFenabled = []
			#self.sg.graph()	
					
					
			TM = {} #dict by tf transition inside slists of lists of start and end states
			
	#compile into our PN representation
	def toPN(self,mod):
		oldtonew = {}
		g = ig.Graph(0,directed=True)
		for node in mod.vs:
			if not node['$type'].find('Place') == -1 or not node['$type'].find('Transition') == -1:
				g.add_vertices(1)
				index = g.vcount()-1
				oldtonew[node.index]=index
				g.vs[index]['name'] = node['name']
				if not node['$type'].find('Place') == -1:
					g.vs[index]['type'] = 'P'
					g.vs[index]['nbTokens'] = node['nbTokens']
				elif not node['$type'].find('Transition') == -1:
					g.vs[index]['type'] = 'T'
					g.vs[index]['fusion'] = node['fusion']
			elif not node['$type'].find('P2T') == -1:
				node['type'] = 'P2T'
			elif not node['$type'].find('T2P') == -1:
				node['type'] = 'T2P'
		
		'''Let's connect'''
		P2T = mod.vs.select(type_eq = 'P2T')
		T2P = mod.vs.select(type_eq = 'T2P')
		for p2t in P2T:
			to = mod.successors(p2t.index)
			fr = mod.predecessors(p2t.index)
			g.add_edges([(oldtonew[fr[0]],oldtonew[to[0]])])
		for t2p in T2P:
			to = mod.successors(t2p.index)
			fr = mod.predecessors(t2p.index)
			g.add_edges([(oldtonew[fr[0]],oldtonew[to[0]])])
		return g
	
	def isReachableFlat(self,state,key=None):
		if not key:
			if self.modules.values()[0].reachable(state):
				return True
			else:
				return False
		else:
			if self.modules[key].reachable(state):
				return True
			else:
				return False

	
	def reachabilityTestModular(self):
		moda = {}
		modb =  {}
		moda['a4'] = 4
		modb['b2'] = 4
		statea = []
		stateb = []
		statea.append(moda)
		stateb.append(modb)
		if not self.isReachableFlat(statea,'a'):
			return False
		if not self.isReachableFlat(stateb,'b'):
			return False
		scca = self.modules['a'].reachableMod(statea)
		print 'A SCC of ancestors %s'%scca
		sccb = self.modules['b'].reachableMod(stateb)
		print 'B SCC of ancestors %s'%sccb
		result = list( list(itertools.product(scca,sccb)))
		
		for node in result:
			v = []
			a = 'a-%d'%node[0]
			b = 'b-%d'%node[1]
			v.append(a)
			v.append(b)
			id = self.sg.statePresentReach(v)
			if not id == -1:
				print 'Modular state space test Found state'
				return True
			
		print 'Modular state space test Not found'
		return False
		
		
		
		
	def reachabilityTestFlat(self):
		moda = {}
		modb =  {}
		moda['a1'] = 2
		modb['b1'] = 2
		state = []
		state.append(moda)
		state.append(modb)
		if self.isReachableFlat(state):
			print 'reachable'
		else:
			print 'Not reachable'
			
	def graph(self,key,g):
		vattr=''
		eattr = ''
		nodes = {}
		graph = pydot.Dot(key, graph_type='digraph')
		dateTag = datetime.datetime.now().strftime("%Y-%b-%d_%H-%M-%S")
		for v in g.vs:
			vattr +='('
			i = len(v['M'])
			for key,value in v['M'].items():
				vattr += '%s-%s'%(key,value)
				if not i-1 == 0:
					vattr+=','
				i -=1
			vattr +=')'
			nodes[v.index] = pydot.Node(vattr)
			graph.add_node(nodes[v.index])
			vattr = ''
		for e in g.es:
			graph.add_edge(pydot.Edge(nodes[e.source],nodes[e.target],label=e['T']))
		graph.write_svg('graphs/STATE%s%s.svg'%(key,dateTag))	
		#graph.write_png('graphs/STATE%s%s.png'%(key,dateTag))	
		
	def graphPN(self,key,g):
		vattr=''
		eattr = ''
		nodes = {}
		graph = pydot.Dot(key, graph_type='digraph')
		dateTag = datetime.datetime.now().strftime("%Y-%b-%d_%H-%M-%S")
		for v in g.vs:
			for at in v.attributes():
				if not v[at] == None:
					vattr += '%s->%s\n'%(at,v[at])
			nodes[v.index] = pydot.Node(vattr)
			graph.add_node(nodes[v.index])
			vattr = ''
		for e in g.es:
			graph.add_edge(pydot.Edge(nodes[e.source],nodes[e.target]))
		graph.write_svg('graphs/PN%s%s.svg'%(key,dateTag))	
		#//COMP 522 project start
